from cschwabpy.models import JSONSerializableBaseModel
from pydantic import ConfigDict, Field
from typing import Mapping, Any, Protocol, Optional
import os
import json
import time
import aiofiles as af
from pathlib import Path

REFRESH_TOKEN_VALIDITY_SECONDS = 7 * 24 * 60 * 60  # 7 days

UNIXTIME_FACTORY = time.time


class Tokens(JSONSerializableBaseModel):
    expires_in: int  # seconds till access__token expires
    token_type: str = "Bearer"
    scope: str
    refresh_token: str
    access_token: str
    id_token: Optional[str] = None
    # rt_created_timestamp: float = Field(default_factory=UNIXTIME_FACTORY)
    created_timestamp: float = Field(default_factory=UNIXTIME_FACTORY)

    @property
    def is_access_token_valid(self) -> bool:
        return time.time() - self.created_timestamp < self.expires_in

    @property
    def is_refresh_token_valid(self) -> bool:
        return time.time() - self.created_timestamp < REFRESH_TOKEN_VALIDITY_SECONDS

    @property
    def all_tokens_invalid(self) -> bool:
        """Whether both RT and AT are invalid."""
        return not self.is_access_token_valid and not self.is_refresh_token_valid


class ITokenStore(Protocol):
    @property
    def token_output_path(self) -> str:
        """Path for outputting tokens."""
        return ""

    def get_tokens(self) -> Optional[Tokens]:
        pass

    def save_tokens(self, tokens: Tokens) -> None:
        pass


class LocalTokenStore(ITokenStore):
    def __init__(
        self, json_file_name: str = "tokens.json", file_path: Optional[str] = None
    ):
        self.file_name = json_file_name
        self.token_file_path = file_path
        if file_path is None:
            self.token_file_path = Path(Path(__file__).parent, json_file_name)
        else:
            self.token_file_path = Path(file_path)

        if not os.path.exists(self.token_file_path.parent):
            os.makedirs(self.token_file_path.parent)

    @property
    def token_output_path(self) -> str:
        return str(self.token_file_path)

    def get_tokens(self) -> Optional[Tokens]:
        try:
            with open(self.token_file_path, "r") as token_file:
                tokens_json = json.loads(token_file.read())
                return Tokens(**tokens_json)
        except:
            return None

    def save_tokens(self, tokens: Tokens) -> None:
        with open(self.token_file_path, "w") as token_file:
            token_file.write(json.dumps(tokens.to_json(), indent=4))


class IAsyncTokenStore(Protocol):
    @property
    def token_output_path(self) -> str:
        """Path for outputting tokens."""
        return ""

    async def get_tokens(self) -> Optional[Tokens]:
        pass

    async def save_tokens(self, tokens: Tokens) -> None:
        pass


class AsyncLocalTokenStore(IAsyncTokenStore):
    def __init__(
        self, json_file_name: str = "tokens.json", file_path: Optional[str] = None
    ):
        self.file_name = json_file_name
        self.token_file_path = file_path
        if file_path is None:
            self.token_file_path = Path(Path(__file__).parent, json_file_name)
        else:
            self.token_file_path = Path(file_path)

        if not os.path.exists(self.token_file_path.parent):
            os.makedirs(self.token_file_path.parent)

    @property
    def token_output_path(self) -> str:
        return str(self.token_file_path)

    async def get_tokens(self) -> Optional[Tokens]:
        try:
            async with af.open(self.token_file_path, mode="r") as token_file:
                token_json_str = await token_file.read()
                tokens_json = json.loads(token_json_str)
                return Tokens(**tokens_json)
        except:
            return None

    async def save_tokens(self, tokens: Tokens) -> None:
        async with af.open(self.token_file_path, mode="w") as token_file:
            await token_file.write(json.dumps(tokens.to_json(), indent=4))



SCHWAB_API_BASE_URL = "https://api.schwabapi.com/v1"
SCHWAB_MARKET_DATA_API_BASE_URL = "https://api.schwabapi.com/marketdata/v1"
SCHWAB_TRADER_API_BASE_URL = "https://api.schwabapi.com/trader/v1"
SCHWAB_AUTH_PATH = "oauth/authorize"
SCHWAB_TOKEN_PATH = "oauth/token"



"""models folder."""
from datetime import datetime, date
from dateutil import parser
from dataclasses import dataclass
from pydantic import BaseModel, ConfigDict, Field
from typing import MutableMapping, Mapping, MutableSet, Any, List, Tuple, Optional
from enum import Enum
import cschwabpy.util as util
import pandas as pd
import pytz

us_eastern_timezone = pytz.timezone("US/Eastern")

OptionChain_Headers = [
    "underlying_price",
    "strike",
    "symbol",
    "lastPrice",
    "openInterest",
    "ask",
    "bid",
    "expiration_date",
    "bid_date",
    "volume",
    "updated_at",
    "gamma",
    "delta",
    "vega",
    "volatility",
]


class JSONSerializableBaseModel(BaseModel):
    model_config = ConfigDict(use_enum_values=True, populate_by_name=True)

    def to_json(self, drop_null_value: bool = True) -> Mapping[str, Any]:
        """Converts the object to a JSON dictionary with options to drop null values from dictionary."""
        super_json = self.model_dump(by_alias=True)
        if not drop_null_value:
            return super_json

        result: MutableMapping[str, Any] = {}
        for k, v in list(super_json.items()):
            if v is not None:
                result[k] = self.__handle_item(v)

        return result

    def __handle_item(self, item: Any) -> Any:
        """Handle item in the dictionary, list or primitive values."""
        if isinstance(item, dict):
            return self.__del_none(item)
        elif isinstance(item, List):
            result = []
            for itm in list(item):
                result.append(self.__handle_item(itm))
            return result
        elif isinstance(item, set):
            result = set()
            for itm in set(item):
                result.add(self.__handle_item(itm))
            return result
        elif isinstance(item, Enum):
            return item.value
        elif isinstance(item, datetime) or isinstance(item, date):
            return str(item)  # because datetime/date object is not JSON serializable
        else:
            return item

    def __del_none(self, d: MutableMapping[str, Any]) -> MutableMapping[str, Any]:
        """Recursively remove None values from a dictionary."""
        for key, value in list(d.items()):
            if value is None:
                del d[key]
            elif isinstance(value, MutableMapping):
                d[key] = self.__del_none(value)

        return d


class ErrorMessage(JSONSerializableBaseModel):
    message: str
    errors: List[str]


class QueryFilterBase(JSONSerializableBaseModel):
    """Base class for query parameters filters."""

    ...

    def to_query_dict(self, ignore_fields: List[str] = []) -> Mapping[str, Any]:
        """Converts the object to a dictionary of query parameters."""
        dump_dict = self.to_json()
        result: MutableMapping[str, Any] = {}
        self_ignore_fields: MutableSet[str] = {}
        for field_key in ignore_fields:
            self_ignore_fields.add(field_key.lower())

        for key, val in dump_dict.items():
            if key.lower() in self_ignore_fields:
                continue
            if val is not None:
                if isinstance(val, Enum):
                    result[key] = val.value
                else:
                    result[key] = val

        return result

    def to_query_params(self, ignore_fields: List[str] = []) -> str:
        """Converts the object to query parameters."""
        query_dict = self.to_query_dict(ignore_fields)
        return "&".join([f"{k}={v}" for k, v in query_dict.items() if v is not None])


class ExpirationType(str, Enum):
    M = "M"
    Q = "Q"
    S = "S"
    W = "W"


class OptionExpiration(JSONSerializableBaseModel):
    expirationDate: str
    daysToExpiration: Optional[int] = None
    expirationType: Optional[ExpirationType] = None
    standard: Optional[bool] = None


class OptionExpirationChainResponse(JSONSerializableBaseModel):
    expirationList: List[OptionExpiration] = []


class OptionContractType(str, Enum):
    """Option contract type."""

    CALL = "CALL"
    PUT = "PUT"
    ALL = "ALL"
    UNKNOWN = "UNKNOWN"


class OptionContractStrategy(str, Enum):
    SINGLE = "SINGLE"
    ANALYTICAL = "ANALYTICAL"
    COVERED = "COVERED"
    VERTICAL = "VERTICAL"
    CALENDAR = "CALENDAR"
    STRANGLE = "STRANGLE"
    STRADDLE = "STRADDLE"
    BUTTERFLY = "BUTTERFLY"
    CONDOR = "CONDOR"
    DIAGONAL = "DIAGONAL"
    COLLAR = "COLLAR"
    ROLL = "ROLL"


class OptionContractRange(str, Enum):
    ITM = "ITM"
    NTM = "NTM"
    OTM = "OTM"
    SAK = "SAK"
    SBK = "SBK"
    SNK = "SNK"
    ALL = "ALL"


class MarketType(str, Enum):
    Equity = "EQUITY"
    Option = "OPTION"
    Future = "FUTURE"
    Bond = "BOND"
    Forex = "FOREX"
    FutureOption = "FUTURE_OPTION"
    Index = "INDEX"
    MutualFund = "MUTUAL_FUND"


class MarketHours(JSONSerializableBaseModel):
    start: str
    end: str

    def open_window(
        self, timezone: pytz.BaseTzInfo = us_eastern_timezone
    ) -> Tuple[datetime, datetime]:
        """Returns the market open window (tuple) in datetime format, defaulted to US eastern timezone.."""
        start_time = parser.parse(self.start).astimezone(timezone)
        end_time = parser.parse(self.end).astimezone(timezone)
        return start_time, end_time


class SessionHours(JSONSerializableBaseModel):
    preMarket: Optional[List[MarketHours]] = None
    regularMarket: List[MarketHours]
    postMarket: Optional[List[MarketHours]] = None


class Market(JSONSerializableBaseModel):
    date: date
    marketType: MarketType
    exchange: Optional[str] = None
    category: Optional[str] = None
    product: str
    productName: str
    isOpen: bool
    sessionHours: Optional[SessionHours] = None


class EquityMarket(JSONSerializableBaseModel):
    EQ: Market


class OptionMarket(JSONSerializableBaseModel):
    EQO: Market
    IND: Market


class MarketHourInfo(JSONSerializableBaseModel):
    equity: Optional[EquityMarket] = None
    option: Optional[OptionMarket] = None

    @property
    def is_equity_market_open(self) -> Optional[bool]:
        if self.equity is None:
            return None

        return self.equity.EQ.isOpen

    @property
    def is_option_market_open(self) -> Optional[bool]:
        if self.option is None:
            return None

        return self.option.EQO.isOpen or self.option.IND.isOpen


class OptionChainQueryFilter(QueryFilterBase):
    symbol: str
    contractType: OptionContractType = OptionContractType.ALL
    strikeCount: int = (
        40  # the number of strikes to  return above or below the at-the-money price
    )
    includeUnderlyingQuote: bool = True
    strategy: OptionContractStrategy = OptionContractStrategy.ANALYTICAL
    interval: Optional[float] = None  # strike interval for spread strategy
    strike: Optional[float] = None  # strike price for spread strategy
    range: Optional[OptionContractRange] = None
    fromDate: str
    toDate: str


class OptionContract(JSONSerializableBaseModel):
    putCall: OptionContractType
    symbol: str
    description: str
    exchangeName: str
    bidPrice: Optional[float] = None
    askPrice: Optional[float] = None
    lastPrice: Optional[float] = None
    markPrice: Optional[float] = None
    bidSize: Optional[int] = None
    askSize: Optional[int] = None
    lastSize: Optional[int] = None
    highPrice: Optional[float] = None
    lowPrice: Optional[float] = None
    openPrice: Optional[float] = None
    closePrice: Optional[float] = None
    totalVolume: Optional[int] = None
    tradeDate: Optional[int] = None
    quoteTimeInLong: Optional[int]
    tradeTimeInLong: Optional[int] = None
    netChange: Optional[float] = None
    volatility: Optional[float] = None
    delta: Optional[float] = None
    gamma: Optional[float] = None
    theta: Optional[float] = None
    vega: Optional[float] = None
    rho: Optional[float] = None
    openInterest: Optional[int] = None
    timeValue: Optional[float] = None
    theoreticalOptionValue: Optional[float] = None
    theoreticalVolatility: Optional[float] = None
    strikePrice: float
    expirationDate: str
    daysToExpiration: int
    expirationType: str  # M for end of month, W for week, Q for quarter, S for third friday
    lastTradingDay: Optional[int] = None
    multiplier: Optional[float] = None
    settlementType: str  # AM, PM
    isIndex: Optional[bool] = None
    percentChange: Optional[float] = None
    markChange: Optional[float] = None
    markPercentChange: Optional[float] = None
    model_config = ConfigDict(
        validate_assignment=False, use_enum_values=True, populate_by_name=True
    )

    def to_dataframe_row(self, strip_space: bool = False) -> List[Any]:
        """Converts the object to a list of values for a dataframe row, strip_space: stripping space in option symbol."""
        symbol = self.symbol.strip().replace(" ", "") if strip_space else self.symbol
        result: List[Any] = [
            self.strikePrice,
            symbol,
            self.lastPrice,
            self.openInterest,
            self.askPrice,
            self.bidPrice,
            self.expirationDate,
            util.ts_to_date_string(self.quoteTimeInLong),
            self.totalVolume,
            util.ts_to_date_string(self.quoteTimeInLong),
            self.gamma,
            self.delta,
            self.vega,
            self.volatility,
        ]
        return result


class Underlying(JSONSerializableBaseModel):
    ask: float
    askSize: int
    bid: float
    bidSize: int
    close_: float = Field(..., alias="close")
    description: Optional[str] = None
    exchangeName: str
    highPrice: Optional[float] = None
    lastPrice: Optional[float] = None
    lowPrice: Optional[float] = None
    mark: Optional[float] = None
    markChange: Optional[float] = None
    markPercentChange: Optional[float] = None
    openPrice: Optional[float] = None
    percentChange: float
    quoteTime: int
    symbol: str
    totalVolume: Optional[int] = None
    tradeTime: Optional[int] = None

    @property
    def quote_time(self) -> Optional[datetime]:
        return util.ts_to_datetime(self.quoteTime)


@dataclass
class OptionChainDataFrames:
    expiration: str
    underlying_symbol: str
    call_df: pd.DataFrame
    put_df: pd.DataFrame


class OptionChain(JSONSerializableBaseModel):
    symbol: str
    status: str
    underlying: Optional[Underlying]
    strategy: str
    interval: Optional[float]
    isDelayed: bool
    isIndex: bool
    interestRate: float
    underlyingPrice: float
    volatility: float
    daysToExpiration: int
    numberOfContracts: int
    putExpDateMap: Mapping[
        str, Mapping[str, List[OptionContract]]
    ]  # key: expiration:27 value:[strike: OptionContract]
    callExpDateMap: Mapping[
        str, Mapping[str, List[OptionContract]]
    ]  # key: expiration:27 value:[strike: OptionContract]

    def to_dataframe_pairs_by_expiration(
        self, strip_space: bool = False
    ) -> List[OptionChainDataFrames]:
        """
        List of OptionChainDataFrames by expiration.
        Each OptionChainDataFrames object contains call and put chain in dataframe format.
        @param strip_space: Whether strip spaces in option symbols.
        """
        results: List[OptionChainDataFrames] = []
        call_map = self.break_down_option_map(
            self.callExpDateMap, strip_space=strip_space
        )
        put_map = self.break_down_option_map(
            self.putExpDateMap, strip_space=strip_space
        )
        for expiration in call_map.keys():
            call_df = call_map[expiration]
            put_df = put_map[expiration]

            cur_df_pair = OptionChainDataFrames(
                expiration=expiration,
                underlying_symbol=self.symbol,
                call_df=call_df,
                put_df=put_df,
            )
            results.append(cur_df_pair)

        return results

    def break_down_option_map(
        self,
        optionExpMap: Mapping[str, Mapping[str, List[OptionContract]]],
        strip_space: bool = False,
    ) -> Mapping[str, pd.DataFrame]:
        """Whether strip spaces in option symbols."""
        result: MutableMapping[str, pd.DataFrame] = {}
        for exp_date, strike_map in optionExpMap.items():
            expiration = exp_date.split(":")[0]
            if expiration not in result:
                result[expiration] = {}

            all_rows = []
            strike_df = pd.DataFrame()
            for strike_str, option_contracts in strike_map.items():
                strike = float(strike_str)
                for option_contract in option_contracts:
                    row = option_contract.to_dataframe_row(strip_space=strip_space)
                    row.insert(0, self.underlying.mark)
                    all_rows.append(row)

                strike_df = pd.DataFrame(all_rows)
                strike_df.columns = OptionChain_Headers
            result[expiration] = strike_df

        return result



from datetime import datetime, date
import pytz
from typing import Optional

eastern_tz: pytz.BaseTzInfo = pytz.timezone("US/Eastern")
YMD_FMT = "%Y-%m-%d"


def now(tz: pytz.BaseTzInfo = eastern_tz) -> datetime:
    """Today datetime now, Returns datetime, default to US East timezone."""  # noqa: DAR201
    if tz is None:
        return datetime.now()
    return datetime.now().astimezone(tz)


def now_unix_ts() -> float:
    """Today now in unix timestamp, Returns timestamp."""  # noqa: DAR201
    return now().timestamp()


def ts_to_datetime(
    ts: Optional[float] = None, tz: pytz.BaseTzInfo = eastern_tz
) -> Optional[datetime]:
    if ts is None:
        return None
    while ts > 1e10:
        ts = ts / 1000
    return datetime.fromtimestamp(ts, tz)


def ts_to_date_string(
    ts: Optional[float] = None, tz: pytz.BaseTzInfo = eastern_tz
) -> Optional[str]:
    if ts is None:
        return None
    return ts_to_datetime(ts, tz).strftime(YMD_FMT)


def to_iso8601_str(dt: datetime) -> str:
    dt = dt.astimezone(eastern_tz)
    dt_str = dt.strftime("%Y-%m-%dT%H:%M:%S")
    return dt_str + ".000Z"


def today_str(tz: pytz.BaseTzInfo = eastern_tz) -> str:  # type: ignore
    """Today in string. Returns Y-m-d."""  # noqa: DAR201
    return now(tz=tz).strftime(YMD_FMT)


def date_to_str(date: date, date_format: str = YMD_FMT) -> str:
    return date.strftime(date_format)



from cschwabpy.models.token import Tokens, ITokenStore, LocalTokenStore
from cschwabpy.models import (
    OptionChainQueryFilter,
    OptionContractType,
    OptionChain,
    OptionExpiration,
    OptionExpirationChainResponse,
    MarketType,
    MarketHourInfo,
)
from cschwabpy.models.trade_models import (
    AccountNumberWithHashID,
    AccountInstrument,
    SecuritiesAccount,
    Account,
    OrderStatus,
    Order,
    InstrumentProjection,
)
import cschwabpy.util as util
import backoff
from datetime import datetime, timedelta
from typing import Optional, List, Mapping
from cschwabpy.costants import (
    SCHWAB_API_BASE_URL,
    SCHWAB_MARKET_DATA_API_BASE_URL,
    SCHWAB_TRADER_API_BASE_URL,
    SCHWAB_AUTH_PATH,
    SCHWAB_TOKEN_PATH,
)

import httpx
import re
import base64
import json

HEADER_ORDER_ID_PATTERN = re.compile(r"orders/(\d+)")


class SchwabClient(object):
    """This is regular sync client. For async client, use SchwabClientAsync."""

    def __init__(
        self,
        app_client_id: str,
        app_secret: str,
        token_store: ITokenStore = LocalTokenStore(),
        tokens: Optional[Tokens] = None,
        http_client: Optional[httpx.Client] = None,
    ) -> None:
        self.__client_id = app_client_id
        self.__client_secret = app_secret
        self.__token_store = token_store
        self.__client = http_client
        self.__keep_client_alive = http_client is not None
        self.__tokens = tokens

    @property
    def token_url(self) -> str:
        return f"{SCHWAB_API_BASE_URL}/{SCHWAB_TOKEN_PATH}"

    @backoff.on_exception(backoff.expo, Exception, max_tries=3, max_time=10)
    def _ensure_valid_access_token(self, force_refresh: bool = False) -> bool:
        if self.__tokens is None:
            self.__tokens = self.__token_store.get_tokens()
        if self.__tokens is None:
            raise Exception(
                "Tokens are not available. Please use get_tokens_manually() to get tokens first."
            )

        if self.__tokens.is_access_token_valid and not force_refresh:
            return True

        client = httpx.Client() if self.__client is None else self.__client
        try:
            key_sec_encoded = self.__encode_app_key_secret()
            response = client.post(
                url=self.token_url,
                headers={
                    "Authorization": f"Basic {key_sec_encoded}",
                    "Content-Type": "application/x-www-form-urlencoded",
                },
                data={
                    "grant_type": "refresh_token",
                    "refresh_token": self.__tokens.refresh_token,
                },
            )

            if response.status_code == 200:
                json_res = response.json()
                self.__tokens = Tokens(**json_res)
                self.__token_store.save_tokens(self.__tokens)
                return True
            else:
                raise Exception(
                    "Status for refreshing access token is not successful. Status: ",
                    response.status_code,
                )
        except Exception as ex:
            print("Failed to refresh access token. Please try again. exception: ", ex)
            return False
        finally:
            if not self.__keep_client_alive:
                client.close()

    def __encode_app_key_secret(self) -> str:
        key_sec = f"{self.__client_id}:{self.__client_secret}"
        return base64.b64encode(key_sec.encode("utf-8")).decode("utf-8")

        # refresh access token
        # doc: https://developer.schwab.com/products/trader-api--individual/details/documentation/Retail%20Trader%20API%20Production

    def __auth_header(self) -> Mapping[str, str]:
        return {
            # "Content-Type": "application/json",
            "Authorization": f"{self.__tokens.token_type} {self.__tokens.access_token}",
            "Accept": "application/json",
        }

    def get_account_numbers(self) -> List[AccountNumberWithHashID]:
        self._ensure_valid_access_token()

        target_url = f"{SCHWAB_TRADER_API_BASE_URL}/accounts/accountNumbers"
        client = httpx.Client() if self.__client is None else self.__client
        try:
            response = client.get(
                url=target_url, params={}, headers=self.__auth_header()
            )
            json_res = response.json()
            account_numbers: List[AccountNumberWithHashID] = []
            for account_json in json_res:
                account_numbers.append(AccountNumberWithHashID(**account_json))
            return account_numbers
        finally:
            if not self.__keep_client_alive:
                client.close()

    def get_accounts(
        self,
        include_positions: bool = True,
        with_account_number_hash: Optional[AccountNumberWithHashID] = None,
    ) -> List[Account]:
        """get all accounts except a specific account_number is provided."""
        self._ensure_valid_access_token()
        target_url = f"{SCHWAB_TRADER_API_BASE_URL}/accounts"
        if with_account_number_hash is not None:
            target_url = f"{target_url}/{with_account_number_hash.hashValue}"

        if include_positions:
            target_url = f"{target_url}?fields=positions"

        client = httpx.Client() if self.__client is None else self.__client
        try:
            response = client.get(
                url=target_url, params={}, headers=self.__auth_header()
            )
            if response.status_code == 200:
                json_res = response.json()
                if with_account_number_hash is None:
                    accounts: List[SecuritiesAccount] = []
                    for account_json in json_res:
                        securities_account = SecuritiesAccount(
                            **account_json
                        ).securitiesAccount
                        accounts.append(securities_account)
                    return accounts
                else:
                    securities_account = SecuritiesAccount(**json_res).securitiesAccount
                    return [securities_account]
            else:
                raise Exception(
                    "Failed to get accounts. Status: ", response.status_code
                )
        finally:
            if not self.__keep_client_alive:
                client.close()

    def get_single_account(
        self,
        with_account_number_hash: AccountNumberWithHashID,
        include_positions: bool = True,
    ) -> Optional[Account]:
        """Convenience method to get a single account by account number's encrypted ID."""
        accounts = self.get_accounts(
            include_positions=include_positions,
            with_account_number_hash=with_account_number_hash,
        )
        if accounts is None or len(accounts) == 0:
            return None

        return accounts[0]

    def get_instruments(
        self,
        symbol: str,
        projection: InstrumentProjection = InstrumentProjection.Fundamental,
    ) -> List[AccountInstrument]:
        self._ensure_valid_access_token()
        target_url = f"{SCHWAB_MARKET_DATA_API_BASE_URL}/instruments?symbol={symbol}&projection={projection.value}"
        client = httpx.Client() if self.__client is None else self.__client

        try:
            response = client.get(
                url=target_url, params={}, headers=self.__auth_header()
            )
            json_res = response.json()
            instruments: List[AccountInstrument] = []
            if "instruments" in json_res:
                for instrument in json_res["instruments"]:
                    instruments.append(AccountInstrument(**instrument))
            return instruments
        finally:
            if not self.__keep_client_alive:
                client.close()

    def cancel_order(
        self, account_number_hash: AccountNumberWithHashID, order_id: int
    ) -> bool:
        """Cancel an order by order ID."""
        self._ensure_valid_access_token()
        target_url = f"{SCHWAB_TRADER_API_BASE_URL}/accounts/{account_number_hash.hashValue}/orders/{order_id}"
        client = httpx.Client() if self.__client is None else self.__client
        try:
            response = client.delete(
                url=target_url, params={}, headers=self.__auth_header()
            )

            return response.status_code == 200
        finally:
            if not self.__keep_client_alive:
                client.close()

    def place_order(
        self, account_number_hash: AccountNumberWithHashID, order: Order
    ) -> int:
        self._ensure_valid_access_token()
        target_url = f"{SCHWAB_TRADER_API_BASE_URL}/accounts/{account_number_hash.hashValue}/orders"
        client = httpx.Client() if self.__client is None else self.__client

        try:
            _header = self.__auth_header()
            _header["Content-Type"] = "application/json"
            response = client.post(
                url=target_url,
                json=json.dumps(order.to_json()),
                headers=_header,
            )
            if response.status_code == 201:
                location_url = response.headers.get("Location")
                if location_url is not None:
                    needle = re.search(HEADER_ORDER_ID_PATTERN, location_url)
                    if needle:
                        return int(needle.group(1))
                    else:
                        raise Exception("Failed to locate order ID in response")

            raise Exception("Failed to place order. Status: ", response.status_code)
        finally:
            if not self.__keep_client_alive:
                client.close()

    def get_order_by_id(
        self,
        account_number_hash: AccountNumberWithHashID,
        order_id: int,
    ) -> Optional[Order]:
        """Get a specific order by order ID."""
        self._ensure_valid_access_token()
        target_url = f"{SCHWAB_TRADER_API_BASE_URL}/accounts/{account_number_hash.hashValue}/orders/{order_id}"
        client = httpx.Client() if self.__client is None else self.__client
        try:
            response = client.get(
                url=target_url, params={}, headers=self.__auth_header()
            )
            if response.status_code == 200:
                order_json = response.json()
                return Order(**order_json)
            elif response.status_code == 404:
                # order not found
                return None
            else:
                raise Exception("Failed to get order. Status: ", response.status_code)
        finally:
            if not self.__keep_client_alive:
                client.close()

    def get_orders(
        self,
        account_number_hash: AccountNumberWithHashID,
        from_entered_time: datetime,
        to_entered_time: datetime,
        max_count: int = 1000,
        status: Optional[OrderStatus] = None,
    ) -> List[Order]:
        self._ensure_valid_access_token()
        target_url = f"{SCHWAB_TRADER_API_BASE_URL}/accounts/{account_number_hash.hashValue}/orders"
        target_url += f"?fromEnteredTime={util.to_iso8601_str(from_entered_time)}&toEnteredTime={util.to_iso8601_str(to_entered_time)}&maxResults={max_count}"
        if status is not None:
            target_url += f"&status={status.value}"

        client = httpx.Client() if self.__client is None else self.__client
        try:
            response = client.get(
                url=target_url, params={}, headers=self.__auth_header()
            )
            if response.status_code == 200:
                json_res = response.json()
                orders: List[Order] = []
                for order_json in json_res:
                    order = Order(**order_json)
                    orders.append(order)
                return orders
            else:
                raise Exception("Failed to get orders. Status: ", response.status_code)
        finally:
            if not self.__keep_client_alive:
                client.close()

    def get_option_expirations(self, underlying_symbol: str) -> List[OptionExpiration]:
        self._ensure_valid_access_token()
        target_url = f"{SCHWAB_MARKET_DATA_API_BASE_URL}/expirationchain?symbol={underlying_symbol}"
        client = httpx.Client() if self.__client is None else self.__client
        try:
            response = client.get(
                url=target_url, params={}, headers=self.__auth_header()
            )
            json_res = response.json()
            expiration_resp = OptionExpirationChainResponse(**json_res)
            return expiration_resp.expirationList
        finally:
            if not self.__keep_client_alive:
                client.close()

    def get_market_hour_info(
        self,
        market_type: Optional[MarketType] = None,
        on_date: Optional[datetime] = None,
    ) -> MarketHourInfo:

        self._ensure_valid_access_token()
        target_url = f"{SCHWAB_MARKET_DATA_API_BASE_URL}/markets"

        if market_type is not None:
            target_url += f"/{market_type.value.lower()}"

        if on_date is not None:
            target_url += f"?date={util.date_to_str(on_date)}"

        client = httpx.Client() if self.__client is None else self.__client
        try:
            response = client.get(
                url=target_url, params={}, headers=self.__auth_header()
            )
            json_res = response.json()
            return MarketHourInfo(**json_res)
        finally:
            if not self.__keep_client_alive:
                client.close()

    def download_option_chain(
        self,
        underlying_symbol: str,
        from_date: str,
        to_date: str,
        contract_type: str = "ALL",
    ) -> OptionChain:
        self._ensure_valid_access_token()

        query_filter = OptionChainQueryFilter(
            symbol=underlying_symbol,
            contractType=OptionContractType(contract_type),
            fromDate=from_date,
            toDate=to_date,
        )
        target_url = (
            f"{SCHWAB_MARKET_DATA_API_BASE_URL}/chains?{query_filter.to_query_params()}"
        )

        client = httpx.Client() if self.__client is None else self.__client
        try:
            response = client.get(
                url=target_url, params={}, headers=self.__auth_header()
            )
            if response.status_code == 200:
                json_res = response.json()
                return OptionChain(**json_res)
            else:
                raise Exception(
                    "Failed to download option chain. Status: ", response.status_code
                )
        finally:
            if not self.__keep_client_alive:
                client.close()

    def get_tokens_manually(
        self,
    ) -> None:
        """Manual steps to get tokens from Charles Schwab API."""
        from prompt_toolkit import prompt
        import urllib.parse as url_parser

        redirect_uri = prompt("Enter your redirect uri> ").strip()
        complete_auth_url = f"{SCHWAB_API_BASE_URL}/{SCHWAB_AUTH_PATH}?response_type=code&client_id={self.__client_id}&redirect_uri={redirect_uri}"
        print(
            f"Copy and open the following URL in browser. Complete Login & Authorization:\n {complete_auth_url}"
        )
        auth_code_response_url = prompt(
            "Paste the entire authorization response URL here> "
        ).strip()

        auth_code = ""
        try:
            auth_code_pattern = re.compile(r"code=(.+)&?")
            d = re.search(auth_code_pattern, auth_code_response_url)
            if d:
                auth_code = d.group(1)
                auth_code = url_parser.unquote(auth_code.split("&")[0])
            else:
                raise Exception(
                    "authorization response url does not contain authorization code"
                )

            if len(auth_code) == 0:
                raise Exception("authorization code is empty")
        except Exception as ex:
            raise Exception(
                "Failed to get authorization code. Please try again. Exception: ", ex
            )

        key_sec_encoded = self.__encode_app_key_secret()
        with httpx.Client() as client:
            response = client.post(
                url=self.token_url,
                headers={
                    "Authorization": f"Basic {key_sec_encoded}",
                    "Content-Type": "application/x-www-form-urlencoded",
                },
                data={
                    "grant_type": "authorization_code",
                    "code": auth_code,
                    "redirect_uri": redirect_uri,
                },
            )

            if response.status_code == 200:
                json_res = response.json()
                tokens = Tokens(**json_res)
                self.__token_store.save_tokens(tokens)
                print(
                    f"Tokens saved successfully at path: {self.__token_store.token_file_path}"
                )
            else:
                print("Failed to get tokens. Please try again.")



from cschwabpy.models import JSONSerializableBaseModel, OptionContractType
from typing import Optional, List, Any, Mapping, MutableMapping
from pydantic import Field
from enum import Enum


class AccountType(str, Enum):
    MARGIN = "MARGIN"
    CASH = "CASH"
    IRA = "IRA"


class OrderStatus(str, Enum):
    AWAITING_PARENT_ORDER = "AWAITING_PARENT_ORDER"
    AWAITING_CONDITION = "AWAITING_CONDITION"
    AWAITING_STOP_CONDITION = "AWAITING_STOP_CONDITION"
    AWAITING_MANUAL_REVIEW = "AWAITING_MANUAL_REVIEW"
    ACCEPTED = "ACCEPTED"
    AWAITING_UR_OUT = "AWAITING_UR_OUT"
    PENDING_ACTIVATION = "PENDING_ACTIVATION"
    QUEUED = "QUEUED"
    WORKING = "WORKING"
    REJECTED = "REJECTED"
    PENDING_CANCEL = "PENDING_CANCEL"
    CANCELED = "CANCELED"
    PENDING_REPLACE = "PENDING_REPLACE"
    REPLACED = "REPLACED"
    FILLED = "FILLED"
    EXPIRED = "EXPIRED"
    NEW = "NEW"
    AWAITING_RELEASE_TIME = "AWAITING_RELEASE_TIME"
    PENDING_ACKNOWLEDGEMENT = "PENDING_ACKNOWLEDGEMENT"
    PENDING_RECALL = "PENDING_RECALL"
    UNKNOWN = "UNKNOWN"


class OrderType(str, Enum):
    MARKET = "MARKET"
    LIMIT = "LIMIT"
    STOP = "STOP"
    STOP_LIMIT = "STOP_LIMIT"
    TRAILING_STOP = "TRAILING_STOP"
    CABINET = "CABINET"
    NON_MARKETABLE = "NON_MARKETABLE"
    MARKET_ON_CLOSE = "MARKET_ON_CLOSE"
    EXERCISE = "EXERCISE"
    TRAILING_STOP_LIMIT = "TRAILING_STOP_LIMIT"
    NET_DEBIT = "NET_DEBIT"
    NET_CREDIT = "NET_CREDIT"
    NET_ZERO = "NET_ZERO"
    LIMIT_ON_CLOSE = "LIMIT_ON_CLOSE"
    UNKNOWN = "UNKNOWN"


class ComplexOrderStrategyType(str, Enum):
    NONE = "NONE"
    COVERED = "COVERED"
    VERTICAL = "VERTICAL"
    BACK_RATIO = "BACK_RATIO"
    CALENDAR = "CALENDAR"
    DIAGONAL = "DIAGONAL"
    STRADDLE = "STRADDLE"
    STRANGLE = "STRANGLE"
    COLLAR_SYNTHETIC = "COLLAR_SYNTHETIC"
    BUTTERFLY = "BUTTERFLY"
    CONDOR = "CONDOR"
    IRON_CONDOR = "IRON_CONDOR"
    VERTICAL_ROLL = "VERTICAL_ROLL"
    COLLAR_WITH_STOCK = "COLLAR_WITH_STOCK"
    DOUBLE_DIAGONAL = "DOUBLE_DIAGONAL"
    UNBALANCED_BUTTERFLY = "UNBALANCED_BUTTERFLY"
    UNBALANCED_CONDOR = "UNBALANCED_CONDOR"
    UNBALANCED_IRON_CONDOR = "UNBALANCED_IRON_CONDOR"
    UNBALANCED_VERTICAL_ROLL = "UNBALANCED_VERTICAL_ROLL"
    MUTUAL_FUND_SWAP = "MUTUAL_FUND_SWAP"
    CUSTOM = "CUSTOM"


class Session(str, Enum):
    NORMAL = "NORMAL"
    AM = "AM"
    PM = "PM"
    SEAMLESS = "SEAMLESS"


class Duration(str, Enum):
    DAY = "DAY"
    GOOD_TILL_CANCEL = "GOOD_TILL_CANCEL"
    FILL_OR_KILL = "FILL_OR_KILL"
    IMMEDIATE_OR_CANCEL = "IMMEDIATE_OR_CANCEL"
    END_OF_WEEK = "END_OF_WEEK"
    END_OF_MONTH = "END_OF_MONTH"
    NEXT_END_OF_MONTH = "NEXT_END_OF_MONTH"
    UNKNOWN = "UNKNOWN"


class Destination(str, Enum):
    INET = "INET"
    ECN_ARCA = "ECN_ARCA"
    CBOE = "CBOE"
    AMEX = "AMEX"
    PHLX = "PHLX"
    ISE = "ISE"
    BOX = "BOX"
    NYSE = "NYSE"
    NASDAQ = "NASDAQ"
    BATS = "BATS"
    C2 = "C2"
    AUTO = "AUTO"


# class OrderLegType(str, Enum):
class AssetType(str, Enum):  # same as OrderLegType
    EQUITY = "EQUITY"
    ETF = "ETF"
    OPTION = "OPTION"
    INDEX = "INDEX"
    MUTUAL_FUND = "MUTUAL_FUND"
    CASH_EQUIVALENT = "CASH_EQUIVALENT"
    FIXED_INCOME = "FIXED_INCOME"
    CURRENCY = "CURRENCY"
    COLLECTIVE_INVESTMENT = "COLLECTIVE_INVESTMENT"


class OptionType(str, Enum):
    VANILLA = "VANILLA"
    BINARY = "BINARY"
    BARRIER = "BARRIER"
    UNKNOWN = "UNKNOWN"


class OrderLegInstruction(str, Enum):
    BUY = "BUY"
    SELL = "SELL"
    BUY_TO_COVER = "BUY_TO_COVER"
    SELL_SHORT = "SELL_SHORT"
    BUY_TO_OPEN = "BUY_TO_OPEN"
    BUY_TO_CLOSE = "BUY_TO_CLOSE"
    SELL_TO_OPEN = "SELL_TO_OPEN"
    SELL_TO_CLOSE = "SELL_TO_CLOSE"
    EXCHANGE = "EXCHANGE"
    SELL_SHORT_EXEMPT = "SELL_SHORT_EXEMPT"


class PositionEffect(str, Enum):
    OPENING = "OPENING"
    CLOSING = "CLOSING"
    AUTOMATIC = "AUTOMATIC"


class SpecialInstruction(str, Enum):
    ALL_OR_NONE = "ALL_OR_NONE"
    DO_NOT_REDUCE = "DO_NOT_REDUCE"
    ALL_OR_NONE_DO_NOT_REDUCE = "ALL_OR_NONE_DO_NOT_REDUCE"


class OrderStrategyType(str, Enum):
    SINGLE = "SINGLE"
    CANCEL = "CANCEL"
    RECALL = "RECALL"
    PAIR = "PAIR"
    FLATTEN = "FLATTEN"
    TWO_DAY_SWAP = "TWO_DAY_SWAP"
    BLAST_ALL = "BLAST_ALL"
    OCO = "OCO"
    TRIGGER = "TRIGGER"


class TaxLotMethod(str, Enum):
    FIFO = "FIFO"
    LIFO = "LIFO"
    HIGH_COST = "HIGH_COST"
    LOW_COST = "LOW_COST"
    AVERAGE_COST = "AVERAGE_COST"
    SPECIFIC_LOT = "SPECIFIC_LOT"
    LOSS_HARVESTER = "LOSS_HARVESTER"


class PriceLinkType(str, Enum):
    VALUE = "VALUE"
    PERCENT = "PERCENT"
    TICK = "TICK"


class StopType(str, Enum):
    STANDARD = "STANDARD"
    BID = "BID"
    ASK = "ASK"
    LAST = "LAST"
    MARK = "MARK"


class PriceLinkBasis(str, Enum):
    MANUAL = "MANUAL"
    BASE = "BASE"
    TRIGGER = "TRIGGER"
    LAST = "LAST"
    BID = "BID"
    ASK = "ASK"
    ASK_BID = "ASK_BID"
    MARK = "MARK"
    AVERAGE = "AVERAGE"


class OrderActivityType(str, Enum):
    EXECUTION = "EXECUTION"
    ORDER_ACTION = "ORDER_ACTION"


class ExecutionType(str, Enum):
    FILL = "FILL"
    CANCELED = "CANCELED"


class QuantityType(str, Enum):
    ALL_SHARES = "ALL_SHARES"
    DOLLARS = "DOLLARS"
    SHARES = "SHARES"


class InstrumentProjection(str, Enum):
    SymbolSearch = "symbol-search"
    SymbolRegex = "symbol-regex"
    DescSearch = "desc-search"
    DescRegex = "desc-regex"
    Search = "search"
    Fundamental = "fundamental"


# --------------------------------------------------


class AccountNumberWithHashID(JSONSerializableBaseModel):
    """accountNumber and hashValue pair.

    Args:
        accountNumber is a numerical string such as 12345678
        hashValue is secret/encrypted ID such as EECD2E7A0B9C1935
        hashValue is used to get resources under the account such as retrieving orders.
    """

    accountNumber: str
    hashValue: str


class MarginBalance(JSONSerializableBaseModel):
    availableFunds: Optional[float] = None
    availableFundsNonMarginableTrade: Optional[float] = None
    buyingPower: Optional[float] = None
    buyingPowerNonMarginableTrade: Optional[float] = None
    dayTradingBuyingPower: Optional[float] = None
    dayTradingBuyingPowerCall: Optional[float] = None
    dayTradingEquityCall: Optional[float] = None
    equity: Optional[float] = None
    equityPercentage: Optional[float] = None
    longMarginValue: Optional[float] = None
    longOptionMarketValue: Optional[float] = None
    longStockValue: Optional[float] = None
    maintenanceCall: Optional[float] = None
    maintenanceRequirement: Optional[float] = None
    margin: Optional[float] = None
    marginEquity: Optional[float] = None
    moneyMarketFund: Optional[float] = None
    mutualFundValue: Optional[float] = None
    sma: Optional[float] = None
    stockBuyingPower: Optional[float] = None
    optionBuyingPower: Optional[float] = None
    regTCall: Optional[float] = None
    shortMarginValue: Optional[float] = None
    shortOptionMarketValue: Optional[float] = None
    shortStockValue: Optional[float] = None
    totalCash: Optional[float] = None
    isInCall: Optional[bool] = None
    unsettledCash: Optional[float] = None
    pendingDeposits: Optional[float] = None
    marginBalance: Optional[float] = None
    shortBalance: Optional[float] = None
    accountValue: Optional[float] = None


class MarginInitialBalance(MarginBalance):
    accruedInterest: Optional[float] = None
    availableFundsNonMarginableTrade: Optional[float] = None
    bondValue: Optional[float] = None
    cashBalance: Optional[float] = None
    cashAvailableForTrading: Optional[float] = None
    cashReceipts: Optional[float] = None
    liquidationValue: Optional[float] = None


class AccountInstrument(JSONSerializableBaseModel):
    assetType: Optional[AssetType] = None
    cusip: Optional[str] = None
    description: Optional[str] = None
    instrumentId: Optional[int] = None
    symbol: Optional[str] = None
    netChange: Optional[float] = None


class AccountEquity(AccountInstrument):
    assetType: AssetType = AssetType.EQUITY


class AccountFixedIncome(AccountInstrument):
    assetType: AssetType = AssetType.FIXED_INCOME
    maturityDate: Optional[str] = None
    factor: Optional[float] = None
    variableRate: Optional[float] = None


class AccountOption(AccountInstrument):
    assetType: AssetType = AssetType.OPTION
    optionDeliverables: Optional[Any] = None  # TODO
    putCall: Optional[OptionContractType] = None
    optionMultiplier: Optional[int] = None
    type_: Optional[OptionType] = Field(None, alias="type")
    underlyingSymbol: Optional[str] = None


class OrderLeg(JSONSerializableBaseModel):
    askPrice: Optional[float] = None
    bidPrice: Optional[float] = None
    lastPrice: Optional[float] = None
    marketPrice: Optional[float] = None
    projectedCommission: Optional[float] = None
    quantity: Optional[float] = None
    finalSymbol: Optional[str] = None
    legId: Optional[int] = None
    assetType: Optional[AssetType] = None
    instruction: Optional[OrderLegInstruction] = None


class OrderLegCollection(JSONSerializableBaseModel):
    orderLegType: Optional[AssetType] = None
    legId: Optional[int] = None
    instrument: Optional[AccountInstrument] = None  # e.g. AccountOption
    instruction: Optional[OrderLegInstruction] = None
    positionEffect: Optional[PositionEffect] = None
    quantity: Optional[float] = None
    quantityType: Optional[QuantityType] = None  # QuantityType.SHARES
    # toSymbol: Optional[str] = None #TODO


class Position(JSONSerializableBaseModel):
    shortQuantity: Optional[float] = None
    averagePrice: Optional[float] = None
    currentDayProfitLoss: Optional[float] = None
    currentDayProfitLossPercentage: Optional[float] = None
    longQuantity: Optional[float] = None
    settledLongQuantity: Optional[float] = None
    settledShortQuantity: Optional[float] = None
    agedQuantity: Optional[float] = None
    instrument: Optional[Any] = None  # TODO: AccountInstrument type
    marketValue: Optional[float] = None
    maintenanceRequirement: Optional[float] = None
    averageLongPrice: Optional[float] = None
    averageShortPrice: Optional[float] = None
    taxLotAverageLongPrice: Optional[float] = None
    taxLotAverageShortPrice: Optional[float] = None
    longOpenProfitLoss: Optional[float] = None
    shortOpenProfitLoss: Optional[float] = None
    previousSessionLongQuantity: Optional[float] = None
    previousSessionShortQuantity: Optional[float] = None
    currentDayCost: Optional[float] = None


class Account(JSONSerializableBaseModel):
    type_: Optional[AccountType] = Field(None, alias="type")
    accountNumber: str
    roundTrips: Optional[int] = None
    isDayTrader: Optional[bool] = False
    isClosingOnlyRestricted: Optional[bool] = False
    pfcbFlag: Optional[bool] = False
    positions: List[Position] = []
    initialBalances: Optional[MarginInitialBalance] = None
    currentBalances: Optional[MarginBalance] = None
    projectedBalances: Optional[MarginBalance] = None

    @property
    def is_margin(self) -> bool:
        return self.type_ == AccountType.MARGIN

    @property
    def is_cash(self) -> bool:
        return self.type_ == AccountType.CASH


class MarginAccount(Account):
    type_: AccountType = Field(AccountType.MARGIN, alias="type")


class CashAccount(Account):
    type_: AccountType = Field(AccountType.CASH, alias="type")


class SecuritiesAccount(JSONSerializableBaseModel):
    securitiesAccount: Account


# Order models
class ExecutionLeg(JSONSerializableBaseModel):
    legId: Optional[int] = 0
    price: Optional[float] = 0
    quantity: Optional[float] = 0
    mismarkedQuantity: Optional[float] = 0
    instrumentId: Optional[int] = 0
    time: Optional[str] = None


class OrderActivity(JSONSerializableBaseModel):
    activityType: Optional[OrderActivityType] = OrderActivityType.ORDER_ACTION
    executionType: Optional[ExecutionType] = ExecutionType.FILL
    quanity: Optional[float] = 0
    orderRemainingQuantity: Optional[float] = 0
    executionLegs: List[ExecutionLeg] = []


class Order(JSONSerializableBaseModel):
    session: Session
    duration: Duration = Duration.DAY
    orderType: OrderType = OrderType.LIMIT
    cancelTime: Optional[str] = None
    complexOrderStrategyType: Optional[ComplexOrderStrategyType] = None
    quantity: Optional[float] = None
    filledQuantity: Optional[float] = None
    remainingQuantity: Optional[float] = None
    requestedDestination: Optional[Destination] = None
    destinationLinkName: Optional[str] = "AUTO"
    releaseTime: Optional[str] = None
    stopPrice: Optional[float] = None
    stopPriceLinkBasis: Optional[PriceLinkBasis] = None
    stopPriceLinkType: Optional[PriceLinkType] = None
    stopPriceOffset: Optional[float] = None
    stopType: Optional[StopType] = None
    priceLinkBasis: Optional[PriceLinkBasis] = None
    priceLinkType: Optional[PriceLinkType] = None
    price: float
    taxLotMethod: Optional[TaxLotMethod] = None
    orderLegCollection: List[OrderLegCollection] = []
    activationPrice: Optional[float] = None
    specialInstruction: Optional[SpecialInstruction] = None
    orderStrategyType: Optional[OrderStrategyType] = OrderStrategyType.SINGLE
    orderId: Optional[int] = None
    cancelable: Optional[bool] = None
    editable: Optional[bool] = None
    status: Optional[OrderStatus] = None
    enteredTime: Optional[str] = None
    closeTime: Optional[str] = None
    tag: Optional[str] = None
    accountNumber: Optional[int] = None
    orderActivityCollection: List[OrderActivity] = []
    replacingOrderCollection: List[str] = []
    childOrderStrategies: List[str] = []
    statusDescription: Optional[str] = None


class EquityOrder(Order):
    quanity: float


class OptionOrder(Order):
    quantity: Optional[float] = None
    complexOrderStrategyType: ComplexOrderStrategyType
